import numpy as np


def whiten(input: np.ndarray) -> np.ndarray:
    r"""Apply whitening (a.k.a sphering).

    Args:
        input (numpy.ndarray):
            Input tensor to be whitened.

    Returns:
        numpy.ndarray of Whitened tensor.
        The type (real or complex) and shape are same as input.

    .. note::
        - If ``input`` is 2D real tensor, it is regarded as (n_channels, n_samples).
        - If ``input`` is 3D complex tensor, it is regarded as (n_channels, n_bins, n_frames).
        - If ``input`` is 3D real tensor, it is regarded as (batch_size, n_channels, n_samples).
        - If ``input`` is 4D complex tensor, it is regarded as
          (batch_size, n_channels, n_bins, n_frames).

    Examples:
        .. code-block:: python

            >>> import numpy as np
            >>> from ssspy.transform import whiten

            >>> n_channels, n_bins, n_frames = 2, 2049, 128
            >>> n_sources = n_channels
            >>> rng = np.random.default_rng(42)

            >>> spectrogram_mix = \
            ...     rng.standard_normal((n_channels, n_bins, n_frames)) \
            ...     + 1j * rng.standard_normal((n_channels, n_bins, n_frames))
            >>> spectrogram_mix_whitened = whiten(spectrogram_mix)
            >>> spectrogram_mix_whitened.shape
            (2, 2049, 128)
    """
    if input.ndim == 2:
        if np.iscomplexobj(input):
            raise ValueError("Real tensor is expected, but given complex tensor.")
        else:
            n_channels = input.shape[0]
            X = input.transpose(1, 0)
            covariance = np.mean(X[:, :, np.newaxis] * X[:, np.newaxis, :], axis=0)
            W, V = np.linalg.eigh(covariance)
            D_diag = 1 / np.sqrt(W)
            D_diag = np.diag(D_diag)
            V_transpose = V.transpose(1, 0)
            output = D_diag @ V_transpose @ X.transpose(1, 0)
    elif input.ndim == 3:
        if np.iscomplexobj(input):
            n_channels = input.shape[0]
            X = input.transpose(1, 2, 0)
            covariance = np.mean(X[:, :, :, np.newaxis] * X[:, :, np.newaxis, :].conj(), axis=1)
            W, V = np.linalg.eigh(covariance)
            D_diag = 1 / np.sqrt(W)
            D_diag = D_diag[:, :, np.newaxis]
            D_diag = D_diag * np.eye(n_channels)
            V_Hermite = V.transpose(0, 2, 1).conj()
            Y = D_diag @ V_Hermite @ X.transpose(0, 2, 1)
            output = Y.transpose(1, 0, 2)
        else:
            n_channels = input.shape[1]
            X = input.transpose(0, 2, 1)
            covariance = np.mean(X[:, :, :, np.newaxis] * X[:, :, np.newaxis, :], axis=1)
            W, V = np.linalg.eigh(covariance)
            D_diag = 1 / np.sqrt(W)
            D_diag = D_diag[:, :, np.newaxis]
            D_diag = D_diag * np.eye(n_channels)
            V_transpose = V.transpose(0, 2, 1)
            output = D_diag @ V_transpose @ X.transpose(0, 2, 1)
    elif input.ndim == 4:
        if np.iscomplexobj(input):
            n_channels = input.shape[1]
            X = input.transpose(0, 2, 3, 1)
            covariance = np.mean(
                X[:, :, :, :, np.newaxis] * X[:, :, :, np.newaxis, :].conj(), axis=2
            )
            W, V = np.linalg.eigh(covariance)
            D_diag = 1 / np.sqrt(W)
            D_diag = D_diag[:, :, :, np.newaxis]
            D_diag = D_diag * np.eye(n_channels)
            V_Hermite = V.transpose(0, 1, 3, 2).conj()
            Y = D_diag @ V_Hermite @ X.transpose(0, 1, 3, 2)
            output = Y.transpose(0, 2, 1, 3)
        else:
            raise ValueError("Complex tensor is expected, but given real tensor.")
    else:
        raise ValueError(
            "The dimension of input is expected 2, 3, or 4, but given {}.".format(input.ndim)
        )

    return output



from typing import Callable, List, Optional, Union

import numpy as np

__all__ = [
    "IterativeMethodBase",
]


class IterativeMethodBase:
    r"""Base class of iterative method.

    This class provides prototype of iterative updates.

    Args:
        callbacks (callable or list[callable], optional):
            Callback functions. Each function is called before separation and at each iteration.
            Default: ``None``.
        record_loss (bool):
            Record the loss at each iteration of the update algorithm if ``record_loss=True``.
            Default: ``True``.
    """

    def __init__(
        self,
        callbacks: Optional[
            Union[
                Callable[["IterativeMethodBase"], None],
                List[Callable[["IterativeMethodBase"], None]],
            ]
        ] = None,
        record_loss: bool = True,
    ) -> None:
        if callbacks is not None:
            if callable(callbacks):
                callbacks = [callbacks]
            self.callbacks = callbacks
        else:
            self.callbacks = None

        self.record_loss = record_loss

        if self.record_loss:
            self.loss = []
        else:
            self.loss = None

    def __call__(self, *args, n_iter: int = 100, initial_call: bool = True, **kwargs) -> np.ndarray:
        r"""Iteratively call ``update_once``.

        Args:
            n_iter (int):
                The number of iterations of demixing filter updates.
                Default: ``100``.
            initial_call (bool):
                If ``True``, perform callbacks (and computation of loss if necessary)
                before iterations.
        """
        if initial_call:
            if self.record_loss:
                loss = self.compute_loss()
                self.loss.append(loss)

            if self.callbacks is not None:
                for callback in self.callbacks:
                    callback(self)

        for _ in range(n_iter):
            self.update_once()

            if self.record_loss:
                loss = self.compute_loss()
                self.loss.append(loss)

            if self.callbacks is not None:
                for callback in self.callbacks:
                    callback(self)

    def update_once(self) -> None:
        r"""Update parameters once."""
        raise NotImplementedError("Implement 'update_once' method.")

    def compute_loss(self) -> float:
        r"""Compute loss.

        Returns:
            Computed loss. The type is expected ``float``.
        """
        raise NotImplementedError("Implement 'compute_loss' method.")



import functools
from typing import Callable, Optional

import numpy as np

from ..special.flooring import identity, max_flooring

EPS = 1e-10


def to_psd(
    X: np.ndarray,
    axis1: int = -2,
    axis2: int = -1,
    flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
        max_flooring, eps=EPS
    ),
) -> np.ndarray:
    r"""Ensure matrix to be positive semidefinite.

    Args:
        X (np.ndarray):
            A complex Hermitian matrix.
        axis1 (int):
            Axis to be used as first axis of 2D sub-arrays.
        axis2 (int):
            Axis to be used as second axis of 2D sub-arrays.
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to return the same shape tensor as the input.
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``functools.partial(max_flooring, eps=1e-10)``.

    Returns:
        Positive semidefinite matrix.
    """
    if flooring_fn is None:
        flooring_fn = identity

    shape = X.shape
    n_dims = len(shape)

    axis1 = n_dims + axis1 if axis1 < 0 else axis1
    axis2 = n_dims + axis2 if axis2 < 0 else axis2

    assert axis1 == n_dims - 2 and axis2 == n_dims - 1, "axis1 == -2 and axis2 == -1"

    if np.iscomplexobj(X):
        X = (X + X.swapaxes(axis1, axis2).conj()) / 2
    else:
        X = (X + X.swapaxes(axis1, axis2)) / 2

    Lamb, P = np.linalg.eigh(X)

    P_Hermite = P.swapaxes(-2, -1)

    if np.iscomplexobj(X):
        P_Hermite = P_Hermite.conj()

    Lamb = flooring_fn(Lamb)
    Lamb = Lamb[..., np.newaxis] * np.eye(Lamb.shape[-1])

    X = P @ Lamb @ P_Hermite

    if np.iscomplexobj(X):
        X = (X + X.swapaxes(axis1, axis2).conj()) / 2
    else:
        X = (X + X.swapaxes(axis1, axis2)) / 2

    return X



from typing import Any, Callable, Optional, Union

import numpy as np

from ..special.flooring import identity


def choose_flooring_fn(
    flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    method: Optional[Any] = None,
) -> Callable[[np.ndarray], np.ndarray]:
    if flooring_fn is None:
        assert method is None, "method is given, but flooring function is not specified."

        flooring_fn = identity
    elif type(flooring_fn) is str and flooring_fn == "self":
        if method is None or not hasattr(method, "flooring_fn"):
            flooring_fn = identity
        else:
            flooring_fn = method.flooring_fn

    assert callable(flooring_fn), "flooring_fn should be callable."

    return flooring_fn



import functools
from typing import Callable, Iterable, List, Optional, Tuple, Union

import numpy as np

from ..algorithm import (
    MINIMAL_DISTORTION_PRINCIPLE_KEYWORDS,
    PROJECTION_BACK_KEYWORDS,
    minimal_distortion_principle,
    projection_back,
)
from ..algorithm.permutation_alignment import correlation_based_permutation_solver
from ..special.flooring import identity, max_flooring
from ..utils.flooring import choose_flooring_fn
from ..utils.select_pair import sequential_pair_selector
from ._update_spatial_model import update_by_ip1, update_by_ip2_one_pair
from .base import IterativeMethodBase

__all__ = [
    "GradFDICA",
    "NaturalGradFDICA",
    "AuxFDICA",
    "GradLaplaceFDICA",
    "NaturalGradLaplaceFDICA",
    "AuxLaplaceFDICA",
]

spatial_algorithms = ["IP", "IP1", "IP2"]
EPS = 1e-10


class FDICABase(IterativeMethodBase):
    r"""Base class of frequency-domain independent component analysis (FDICA).

    Args:
        contrast_fn (callable):
            A contrast function which corresponds to :math:`-\log p(y_{ijn})`.
            This function is expected to receive (n_channels, n_bins, n_frames)
            and return (n_channels, n_bins, n_frames).
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to receive (n_channels, n_bins, n_frames)
            and return (n_channels, n_bins, n_frames).
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``partial(max_flooring, eps=1e-10)``.
        callbacks (callable or list[callable], optional):
            Callback functions. Each function is called before separation and at each iteration.
            Default: ``None``.
        permutation_alignment (bool):
            If ``permutation_alignment=True``, a permutation solver is used to align
            estimated spectrograms. Default: ``True``.
        scale_restoration (bool or str):
            Technique to restore scale ambiguity.
            If ``scale_restoration=True``, the projection back technique is applied to
            estimated spectrograms. You can also specify ``projection_back``
            or ``minimal_distortion_principle``. Default: ``True``.
        record_loss (bool):
            Record the loss at each iteration of the update algorithm if ``record_loss=True``.
            Default: ``True``.
        reference_id (int):
            Reference channel for projection back and minimal distortion principle. Default: ``0``.
    """

    def __init__(
        self,
        contrast_fn: Callable[[np.ndarray], np.ndarray] = None,
        flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
            max_flooring, eps=EPS
        ),
        callbacks: Optional[
            Union[Callable[["FDICABase"], None], List[Callable[["FDICABase"], None]]]
        ] = None,
        permutation_alignment: bool = True,
        scale_restoration: Union[bool, str] = True,
        record_loss: bool = True,
        reference_id: int = 0,
    ) -> None:
        super().__init__(callbacks=callbacks, record_loss=record_loss)

        if contrast_fn is None:
            raise ValueError("Specify contrast function.")
        else:
            self.contrast_fn = contrast_fn

        if flooring_fn is None:
            self.flooring_fn = identity
        else:
            self.flooring_fn = flooring_fn

        self.input = None
        self.permutation_alignment = permutation_alignment
        self.scale_restoration = scale_restoration

        if reference_id is None and scale_restoration:
            raise ValueError("Specify 'reference_id' if scale_restoration=True.")
        else:
            self.reference_id = reference_id

    def __call__(
        self, input: np.ndarray, n_iter: int = 100, initial_call: bool = True, **kwargs
    ) -> np.ndarray:
        r"""Separate a frequency-domain multichannel signal.

        Args:
            input (numpy.ndarray):
                Mixture signal in frequency-domain.
                The shape is (n_channels, n_bins, n_frames).
            n_iter (int):
                Number of iterations of demixing filter updates.
                Default: ``100``.
            initial_call (bool):
                If ``True``, perform callbacks (and computation of loss if necessary)
                before iterations.

        Returns:
            numpy.ndarray of the separated signal in frequency-domain.
            The shape is (n_channels, n_bins, n_frames).
        """
        self.input = input.copy()

        self._reset(**kwargs)

        super().__call__(n_iter=n_iter, initial_call=initial_call)

        raise NotImplementedError("Implement '__call__' method.")

    def __repr__(self) -> str:
        s = "FDICA("
        s += ", permutation_alignment={permutation_alignment}"
        s += ", scale_restoration={scale_restoration}"
        s += ", record_loss={record_loss}"

        if self.scale_restoration:
            s += ", reference_id={reference_id}"

        s += ")"

        return s.format(**self.__dict__)

    def _reset(self, **kwargs) -> None:
        r"""Reset attributes by given keyword arguments.

        Args:
            kwargs:
                Keyword arguments to set as attributes of FDICA.
        """
        assert self.input is not None, "Specify data!"

        for key in kwargs.keys():
            setattr(self, key, kwargs[key])

        X = self.input

        n_channels, n_bins, n_frames = X.shape
        n_sources = n_channels  # n_channels == n_sources

        self.n_sources, self.n_channels = n_sources, n_channels
        self.n_bins, self.n_frames = n_bins, n_frames

        if not hasattr(self, "demix_filter"):
            W = np.eye(n_sources, n_channels, dtype=np.complex128)
            W = np.tile(W, reps=(n_bins, 1, 1))
        else:
            if self.demix_filter is None:
                W = None
            else:
                # To avoid overwriting ``demix_filter`` given by keyword arguments.
                W = self.demix_filter.copy()

        self.demix_filter = W
        self.output = self.separate(X, demix_filter=W)

    def separate(self, input: np.ndarray, demix_filter: np.ndarray) -> np.ndarray:
        r"""Separate ``input`` using ``demixing_filter``.

        .. math::
            \boldsymbol{y}_{ij}
            = \boldsymbol{W}_{i}\boldsymbol{x}_{ij}

        Args:
            input (numpy.ndarray):
                The mixture signal in frequency-domain.
                The shape is (n_channels, n_bins, n_frames).
            demix_filter (numpy.ndarray):
                The demixing filters to separate ``input``.
                The shape is (n_bins, n_sources, n_channels).

        Returns:
            numpy.ndarray of the separated signal in frequency-domain.
            The shape is (n_sources, n_bins, n_frames).
        """
        X, W = input, demix_filter
        Y = W @ X.transpose(1, 0, 2)
        output = Y.transpose(1, 0, 2)

        return output

    def compute_loss(self) -> float:
        r"""Compute loss :math:`\mathcal{L}`.

        :math:`\mathcal{L}` is given as follows:

        .. math::
            \mathcal{L}
            &= \sum_{i}\mathcal{L}^{[i]}, \\
            \mathcal{L}^{[i]}
            &= \frac{1}{J}\sum_{j,n}G(y_{ijn})
            - 2\log|\det\boldsymbol{W}_{i}|, \\
            G(y_{ijn}) \
            &= - \log p(y_{ijn})

        Returns:
            Computed loss.
        """
        X, W = self.input, self.demix_filter
        Y = self.separate(X, demix_filter=W)  # (n_sources, n_bins, n_frames)
        logdet = self.compute_logdet(W)  # (n_bins,)
        G = self.contrast_fn(Y)  # (n_sources, n_bins, n_frames)
        loss = np.sum(np.mean(G, axis=2), axis=0) - 2 * logdet
        loss = loss.sum(axis=0).item()

        return loss

    def compute_logdet(self, demix_filter: np.ndarray) -> np.ndarray:
        r"""Compute log-determinant of demixing filter.

        Args:
            demix_filter (numpy.ndarray):
                Demixing filters with shape of (n_bins, n_sources, n_channels).

        Returns:
            numpy.ndarray of computed log-determinant values.
        """
        _, logdet = np.linalg.slogdet(demix_filter)  # (n_bins,)

        return logdet

    def solve_permutation(self) -> None:
        r"""Align demixing filters and separated spectrograms"""

        permutation_alignment = self.permutation_alignment

        assert permutation_alignment, "Set permutation_alignment=True."

        if type(permutation_alignment) is bool:
            # when permutation_alignment is True
            permutation_alignment = "spectrogram_correlation"

        if permutation_alignment == "spectrogram_correlation":
            self.solve_permutation_by_correlation()
        else:
            raise NotImplementedError(
                "permutation_alignment {} is not implemented.".format(permutation_alignment)
            )

    def solve_permutation_by_correlation(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Align posteriors and separated spectrograms by correlation.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.
        """

        flooring_fn = choose_flooring_fn(flooring_fn, method=self)
        X, W = self.input, self.demix_filter

        Y = self.separate(X, demix_filter=W)
        Y = Y.transpose(1, 0, 2)
        Y, W = correlation_based_permutation_solver(Y, W, flooring_fn=flooring_fn)
        Y = Y.transpose(1, 0, 2)

        self.output, self.demix_filter = Y, W

    def restore_scale(self) -> None:
        r"""Restore scale ambiguity.

        If ``self.scale_restoration=projection_back``, we use projection back technique.
        If ``self.scale_restoration=minimal_distortion_principle``,
        we use minimal distortion principle.
        """
        scale_restoration = self.scale_restoration

        assert scale_restoration, "Set self.scale_restoration=True."

        if type(scale_restoration) is bool:
            scale_restoration = PROJECTION_BACK_KEYWORDS[0]

        if scale_restoration in PROJECTION_BACK_KEYWORDS:
            self.apply_projection_back()
        elif scale_restoration in MINIMAL_DISTORTION_PRINCIPLE_KEYWORDS:
            self.apply_minimal_distortion_principle()
        else:
            raise ValueError("{} is not supported for scale restoration.".format(scale_restoration))

    def apply_projection_back(self) -> None:
        r"""Apply projection back technique to estimated spectrograms."""
        assert self.scale_restoration, "Set self.scale_restoration=True."

        X, W = self.input, self.demix_filter
        W_scaled = projection_back(W, reference_id=self.reference_id)
        Y_scaled = self.separate(X, demix_filter=W_scaled)

        self.output, self.demix_filter = Y_scaled, W_scaled

    def apply_minimal_distortion_principle(self) -> None:
        r"""Apply minimal distortion principle to estimated spectrograms."""
        assert self.scale_restoration, "Set self.scale_restoration=True."

        X, W = self.input, self.demix_filter
        Y = self.separate(X, demix_filter=W)
        Y_scaled = minimal_distortion_principle(Y, reference=X, reference_id=self.reference_id)
        X = X.transpose(1, 0, 2)
        Y = Y_scaled.transpose(1, 0, 2)
        X_Hermite = X.transpose(0, 2, 1).conj()
        W_scaled = Y @ X_Hermite @ np.linalg.inv(X @ X_Hermite)

        self.output, self.demix_filter = Y_scaled, W_scaled


class GradFDICABase(FDICABase):
    r"""Base class of frequency-domain independent component analysis (FDICA) \
    using the gradient descent.

    Args:
        step_size (float):
            A step size of the gradient descent. Default: ``1e-1``.
        contrast_fn (callable):
            A contrast function which corresponds to :math:`-\log p(y_{ijn})`.
            This function is expected to receive (n_channels, n_bins, n_frames)
            and return (n_channels, n_bins, n_frames).
        score_fn (callable):
            A score function which corresponds to the partial derivative of the contrast function.
            This function is expected to receive (n_channels, n_bins, n_frames)
            and return (n_channels, n_bins, n_frames).
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to receive (n_channels, n_bins, n_frames)
            and return (n_channels, n_bins, n_frames).
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``partial(max_flooring, eps=1e-10)``.
        callbacks (callable or list[callable], optional):
            Callback functions. Each function is called before separation and at each iteration.
            Default: ``None``.
        permutation_alignment (bool):
            If ``permutation_alignment=True``, a permutation solver is used to align
            estimated spectrograms. Default: ``True``.
        scale_restoration (bool or str):
            Technique to restore scale ambiguity.
            If ``scale_restoration=True``, the projection back technique is applied to
            estimated spectrograms. You can also specify ``projection_back``
            or ``minimal_distortion_principle``. Default: ``True``.
        record_loss (bool):
            Record the loss at each iteration of the gradient descent if ``record_loss=True``.
            Default: ``True``.
        reference_id (int):
            Reference channel for projection back and minimal distortion principle. Default: ``0``.
    """

    def __init__(
        self,
        step_size: float = 1e-1,
        contrast_fn: Callable[[np.ndarray], np.ndarray] = None,
        score_fn: Callable[[np.ndarray], np.ndarray] = None,
        flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
            max_flooring, eps=EPS
        ),
        callbacks: Optional[
            Union[Callable[["GradFDICABase"], None], List[Callable[["GradFDICABase"], None]]]
        ] = None,
        permutation_alignment: bool = True,
        scale_restoration: Union[bool, str] = True,
        record_loss: bool = True,
        reference_id: int = 0,
    ) -> None:
        super().__init__(
            contrast_fn=contrast_fn,
            flooring_fn=flooring_fn,
            callbacks=callbacks,
            permutation_alignment=permutation_alignment,
            scale_restoration=scale_restoration,
            record_loss=record_loss,
            reference_id=reference_id,
        )

        self.step_size = step_size

        if score_fn is None:
            raise ValueError("Specify score function.")
        else:
            self.score_fn = score_fn

    def __call__(
        self, input: np.ndarray, n_iter: int = 100, initial_call: bool = True, **kwargs
    ) -> np.ndarray:
        r"""Separate a frequency-domain multichannel signal.

        Args:
            input (numpy.ndarray):
                The mixture signal in frequency-domain.
                The shape is (n_channels, n_bins, n_frames).
            n_iter (int):
                The number of iterations of demixing filter updates.
                Default: ``100``.
            initial_call (bool):
                If ``True``, perform callbacks (and computation of loss if necessary)
                before iterations.

        Returns:
            numpy.ndarray of the separated signal in frequency-domain.
            The shape is (n_channels, n_bins, n_frames).
        """
        self.input = input.copy()

        self._reset(**kwargs)

        # Call __call__ of FDICABase's parent, i.e. __call__ of IterativeMethodBase
        super(FDICABase, self).__call__(n_iter=n_iter, initial_call=initial_call)

        if self.permutation_alignment:
            self.solve_permutation()

        if self.scale_restoration:
            self.restore_scale()

        self.output = self.separate(self.input, demix_filter=self.demix_filter)

        return self.output

    def __repr__(self) -> str:
        s = "GradFDICA("
        s += "step_size={step_size}"
        s += ", permutation_alignment={permutation_alignment}"
        s += ", scale_restoration={scale_restoration}"
        s += ", record_loss={record_loss}"

        if self.scale_restoration:
            s += ", reference_id={reference_id}"

        s += ")"

        return s.format(**self.__dict__)

    def update_once(self) -> None:
        r"""Update demixing filters once."""
        raise NotImplementedError("Implement 'update_once' method.")


class GradFDICA(GradFDICABase):
    r"""Frequency-domain independent component analysis (FDICA) \
    using the gradient descent.

    Args:
        step_size (float):
            A step size of the gradient descent. Default: ``1e-1``.
        contrast_fn (callable):
            A contrast function corresponds to :math:`-\log p(y_{ijn})`.
            This function is expected to receive (n_channels, n_bins, n_frames)
            and return (n_channels, n_bins, n_frames).
        score_fn (callable):
            A score function corresponds to the partial derivative of the contrast function.
            This function is expected to receive (n_channels, n_bins, n_frames)
            and return (n_channels, n_bins, n_frames).
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to receive (n_channels, n_bins, n_frames)
            and return (n_channels, n_bins, n_frames).
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``partial(max_flooring, eps=1e-10)``.
        callbacks (callable or list[callable], optional):
            Callback functions. Each function is called before separation and at each iteration.
            Default: ``None``.
        is_holonomic (bool):
            If ``is_holonomic=True``, Holonomic-type update is used.
            Otherwise, Nonholonomic-type update is used. Default: ``False``.
        permutation_alignment (bool):
            If ``permutation_alignment=True``, a permutation solver is used to align
            estimated spectrograms. Default: ``True``.
        scale_restoration (bool or str):
            Technique to restore scale ambiguity.
            If ``scale_restoration=True``, the projection back technique is applied to
            estimated spectrograms. You can also specify ``projection_back``
            or ``minimal_distortion_principle``. Default: ``True``.
        record_loss (bool):
            Record the loss at each iteration of the gradient descent if ``record_loss=True``.
            Default: ``True``.
        reference_id (int):
            Reference channel for projection back and minimal distortion principle. Default: ``0``.

    Examples:
        Update demixing filters using Holonomic-type update:

        .. code-block:: python

            >>> def contrast_fn(y):
            ...     return 2 * np.abs(y)

            >>> def score_fn(y):
            ...     denom = np.maximum(np.abs(y), 1e-10)
            ...     return y / denom

            >>> n_channels, n_bins, n_frames = 2, 2049, 128
            >>> spectrogram_mix = \
            ...     np.random.randn(n_channels, n_bins, n_frames) \
            ...     + 1j * np.random.randn(n_channels, n_bins, n_frames)

            >>> fdica = GradFDICA(
            ...     contrast_fn=contrast_fn,
            ...     score_fn=score_fn,
            ...     is_holonomic=True,
            ... )
            >>> spectrogram_est = fdica(spectrogram_mix, n_iter=1000)
            >>> print(spectrogram_mix.shape, spectrogram_est.shape)
            (2, 2049, 128), (2, 2049, 128)

        Update demixing filters using Nonholonomic-type update:

        .. code-block:: python

            >>> def contrast_fn(y):
            ...     return 2 * np.abs(y)

            >>> def score_fn(y):
            ...     denom = np.maximum(np.abs(y), 1e-10)
            ...     return y / denom

            >>> n_channels, n_bins, n_frames = 2, 2049, 128
            >>> spectrogram_mix = \
            ...     np.random.randn(n_channels, n_bins, n_frames) \
            ...     + 1j * np.random.randn(n_channels, n_bins, n_frames)

            >>> fdica = GradFDICA(
            ...     contrast_fn=contrast_fn,
            ...     score_fn=score_fn,
            ...     is_holonomic=False,
            ... )
            >>> spectrogram_est = fdica(spectrogram_mix, n_iter=1000)
            >>> print(spectrogram_mix.shape, spectrogram_est.shape)
            (2, 2049, 128), (2, 2049, 128)
    """

    def __init__(
        self,
        step_size: float = 1e-1,
        contrast_fn: Callable[[np.ndarray], np.ndarray] = None,
        score_fn: Callable[[np.ndarray], np.ndarray] = None,
        flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
            max_flooring, eps=EPS
        ),
        callbacks: Optional[
            Union[Callable[["GradFDICA"], None], List[Callable[["GradFDICA"], None]]]
        ] = None,
        is_holonomic: bool = False,
        permutation_alignment: bool = True,
        scale_restoration: Union[bool, str] = True,
        record_loss: bool = True,
        reference_id: int = 0,
    ) -> None:
        super().__init__(
            step_size=step_size,
            contrast_fn=contrast_fn,
            score_fn=score_fn,
            flooring_fn=flooring_fn,
            callbacks=callbacks,
            permutation_alignment=permutation_alignment,
            scale_restoration=scale_restoration,
            record_loss=record_loss,
            reference_id=reference_id,
        )

        self.is_holonomic = is_holonomic

    def __repr__(self) -> str:
        s = "GradFDICA("
        s += "step_size={step_size}"
        s += ", is_holonomic={is_holonomic}"
        s += ", permutation_alignment={permutation_alignment}"
        s += ", scale_restoration={scale_restoration}"
        s += ", record_loss={record_loss}"

        if self.scale_restoration:
            s += ", reference_id={reference_id}"

        s += ")"

        return s.format(**self.__dict__)

    def update_once(self) -> None:
        r"""Update demixing filters once using the gradient descent.

        If ``is_holonomic=True``, demixing filters are updated as follows:

        .. math::
            \boldsymbol{W}_{i}
            \leftarrow\boldsymbol{W}_{i} - \eta\left(\frac{1}{J}\sum_{j}
            \boldsymbol{\phi}(\boldsymbol{y}_{ij})\boldsymbol{y}_{ij}^{\mathsf{H}}
            -\boldsymbol{I}\right)\boldsymbol{W}_{i}^{-\mathsf{H}},

        where

        .. math::
            \boldsymbol{\phi}(\boldsymbol{y}_{ij})
            &= \left(\phi(y_{ij1}),\ldots,\phi(y_{ijn}),\ldots,\phi(y_{ijN})
            \right)^{\mathsf{T}}\in\mathbb{C}^{N}, \\
            \phi(y_{ijn})
            &= \frac{\partial G(y_{ijn})}{\partial y_{ijn}^{*}}, \\
            G(y_{ijn})
            &= -\log p(y_{ijn}).

        Otherwise (``is_holonomic=False``),

        .. math::
            \boldsymbol{W}_{i}
            \leftarrow\boldsymbol{W}_{i}
            - \eta\cdot\mathrm{offdiag}\left(\frac{1}{J}\sum_{j}
            \boldsymbol{\phi}(\boldsymbol{y}_{ij})\boldsymbol{y}_{ij}^{\mathsf{H}}\right)
            \boldsymbol{W}_{i}^{-\mathsf{H}}.
        """
        X, W = self.input, self.demix_filter
        Y = self.separate(X, demix_filter=W)

        Phi = self.score_fn(Y)
        Y_conj = Y.conj()
        PhiY = np.mean(Phi[:, np.newaxis, :, :] * Y_conj[np.newaxis, :, :, :], axis=-1)
        PhiY = PhiY.transpose(2, 0, 1)  # (n_bins, n_sources, n_sources)
        W_inv = np.linalg.inv(W)
        W_inv_Hermite = W_inv.transpose(0, 2, 1).conj()
        eye = np.eye(self.n_sources)

        if self.is_holonomic:
            delta = (PhiY - eye) @ W_inv_Hermite
        else:
            delta = ((1 - eye) * PhiY) @ W_inv_Hermite

        W = W - self.step_size * delta

        Y = self.separate(X, demix_filter=W)

        self.demix_filter = W
        self.output = Y


class NaturalGradFDICA(GradFDICABase):
    r"""Frequency-domain independent component analysis (FDICA) \
    using the natural gradient descent.

    Args:
        step_size (float):
            A step size of the gradient descent. Default: ``1e-1``.
        contrast_fn (callable):
            A contrast function corresponds to :math:`-\log p(y_{ijn})`.
            This function is expected to receive (n_channels, n_bins, n_frames)
            and return (n_channels, n_bins, n_frames).
        score_fn (callable):
            A score function corresponds to the partial derivative of the contrast function.
            This function is expected to receive (n_channels, n_bins, n_frames)
            and return (n_channels, n_bins, n_frames).
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to receive (n_channels, n_bins, n_frames)
            and return (n_channels, n_bins, n_frames).
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``partial(max_flooring, eps=1e-10)``.
        callbacks (callable or list[callable], optional):
            Callback functions. Each function is called before separation and at each iteration.
            Default: ``None``.
        is_holonomic (bool):
            If ``is_holonomic=True``, Holonomic-type update is used.
            Otherwise, Nonholonomic-type update is used. Default: ``False``.
        permutation_alignment (bool):
            If ``permutation_alignment=True``, a permutation solver is used to align
            estimated spectrograms. Default: ``True``.
        scale_restoration (bool or str):
            Technique to restore scale ambiguity.
            If ``scale_restoration=True``, the projection back technique is applied to
            estimated spectrograms. You can also specify ``projection_back``
            or ``minimal_distortion_principle``. Default: ``True``.
        record_loss (bool):
            Record the loss at each iteration of the gradient descent if ``record_loss=True``.
            Default: ``True``.
        reference_id (int):
            Reference channel for projection back and minimal distortion principle. Default: ``0``.

    Examples:
        Update demixing filters using Holonomic-type update:

        .. code-block:: python

            >>> def contrast_fn(y):
            ...     return 2 * np.abs(y)

            >>> def score_fn(y):
            ...     denom = np.maximum(np.abs(y), 1e-10)
            ...     return y / denom

            >>> n_channels, n_bins, n_frames = 2, 2049, 128
            >>> spectrogram_mix = \
            ...     np.random.randn(n_channels, n_bins, n_frames) \
            ...     + 1j * np.random.randn(n_channels, n_bins, n_frames)

            >>> fdica = NaturalGradFDICA(
            ...     contrast_fn=contrast_fn,
            ...     score_fn=score_fn,
            ...     is_holonomic=True,
            ... )
            >>> spectrogram_est = fdica(spectrogram_mix, n_iter=1000)
            >>> print(spectrogram_mix.shape, spectrogram_est.shape)
            (2, 2049, 128), (2, 2049, 128)

        Update demixing filters using Nonholonomic-type update:

        .. code-block:: python

            >>> def contrast_fn(y):
            ...     return 2 * np.abs(y)

            >>> def score_fn(y):
            ...     denom = np.maximum(np.abs(y), 1e-10)
            ...     return y / denom

            >>> n_channels, n_bins, n_frames = 2, 2049, 128
            >>> spectrogram_mix = \
            ...     np.random.randn(n_channels, n_bins, n_frames) \
            ...     + 1j * np.random.randn(n_channels, n_bins, n_frames)

            >>> fdica = NaturalGradFDICA(
            ...     contrast_fn=contrast_fn,
            ...     score_fn=score_fn,
            ...     is_holonomic=False,
            ... )
            >>> spectrogram_est = fdica(spectrogram_mix, n_iter=1000)
            >>> print(spectrogram_mix.shape, spectrogram_est.shape)
            (2, 2049, 128), (2, 2049, 128)
    """

    def __init__(
        self,
        step_size: float = 1e-1,
        contrast_fn: Callable[[np.ndarray], np.ndarray] = None,
        score_fn: Callable[[np.ndarray], np.ndarray] = None,
        flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
            max_flooring, eps=EPS
        ),
        callbacks: Optional[
            Union[Callable[["NaturalGradFDICA"], None], List[Callable[["NaturalGradFDICA"], None]]]
        ] = None,
        is_holonomic: bool = False,
        permutation_alignment: bool = True,
        scale_restoration: Union[bool, str] = True,
        record_loss: bool = True,
        reference_id: int = 0,
    ) -> None:
        super().__init__(
            step_size=step_size,
            contrast_fn=contrast_fn,
            score_fn=score_fn,
            flooring_fn=flooring_fn,
            callbacks=callbacks,
            permutation_alignment=permutation_alignment,
            scale_restoration=scale_restoration,
            record_loss=record_loss,
            reference_id=reference_id,
        )

        self.is_holonomic = is_holonomic

    def __repr__(self) -> str:
        s = "NaturalGradFDICA("
        s += "step_size={step_size}"
        s += ", is_holonomic={is_holonomic}"
        s += ", permutation_alignment={permutation_alignment}"
        s += ", scale_restoration={scale_restoration}"
        s += ", record_loss={record_loss}"

        if self.scale_restoration:
            s += ", reference_id={reference_id}"

        s += ")"

        return s.format(**self.__dict__)

    def update_once(self) -> None:
        r"""Update demixing filters once using the gradient descent.

        If ``is_holonomic=True``, demixing filters are updated as follows:

        .. math::
            \boldsymbol{W}_{i}
            \leftarrow\boldsymbol{W}_{i} - \eta\left(\frac{1}{J}\sum_{j}
            \boldsymbol{\phi}(\boldsymbol{y}_{ij})\boldsymbol{y}_{ij}^{\mathsf{H}}
            -\boldsymbol{I}\right)\boldsymbol{W}_{i},

        where

        .. math::
            \boldsymbol{\phi}(\boldsymbol{y}_{ij})
            &= \left(\phi(y_{ij1}),\ldots,\phi(y_{ijn}),\ldots,\phi(y_{ijN})
            \right)^{\mathsf{T}}\in\mathbb{C}^{N}, \\
            \phi(y_{ijn})
            &= \frac{\partial G(y_{ijn})}{\partial y_{ijn}^{*}}, \\
            G(y_{ijn})
            &= -\log p(y_{ijn}).

        Otherwise (``is_holonomic=False``),

        .. math::
            \boldsymbol{W}_{i}
            \leftarrow\boldsymbol{W}_{i}
            - \eta\cdot\mathrm{offdiag}\left(\frac{1}{J}\sum_{j}
            \boldsymbol{\phi}(\boldsymbol{y}_{ij})\boldsymbol{y}_{ij}^{\mathsf{H}}\right)
            \boldsymbol{W}_{i}.
        """
        X, W = self.input, self.demix_filter
        Y = self.separate(X, demix_filter=W)

        Phi = self.score_fn(Y)
        Y_conj = Y.conj()
        PhiY = np.mean(Phi[:, np.newaxis, :, :] * Y_conj[np.newaxis, :, :, :], axis=-1)
        PhiY = PhiY.transpose(2, 0, 1)  # (n_bins, n_sources, n_sources)
        eye = np.eye(self.n_sources)

        if self.is_holonomic:
            delta = (PhiY - eye) @ W
        else:
            delta = ((1 - eye) * PhiY) @ W

        W = W - self.step_size * delta

        Y = self.separate(X, demix_filter=W)

        self.demix_filter = W
        self.output = Y


class AuxFDICA(FDICABase):
    r"""Auxiliary-function-based frequency-domain independent component analysis \
    (AuxFDICA) [#ono2010auxiliary]_.

    Args:
        spatial_algorithm (str):
            Algorithm to update demixing filters.
            Choose ``IP``, ``IP1``, or ``IP2``. Default: ``IP``.
        contrast_fn (callable):
            A contrast function corresponds to :math:`-\log p(y_{ijn})`.
            This function is expected to receive (n_channels, n_bins, n_frames)
            and return (n_channels, n_bins, n_frames).
        d_contrast_fn (callable):
            A partial derivative of the real contrast function.
            This function is expected to receive (n_channels, n_bins, n_frames)
            and return (n_channels, n_bins, n_frames).
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to receive (n_channels, n_bins, n_frames)
            and return (n_channels, n_bins, n_frames).
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``partial(max_flooring, eps=1e-10)``.
        pair_selector (callable, optional):
            Selector to choose updaing pair in ``IP2`` and ``ISS2``.
            If ``None`` is given, ``partial(sequential_pair_selector, sort=True)`` is used.
            Default: ``None``.
        callbacks (callable or list[callable], optional):
            Callback functions. Each function is called before separation and at each iteration.
            Default: ``None``.
        permutation_alignment (bool):
            If ``permutation_alignment=True``, a permutation solver is used to align
            estimated spectrograms. Default: ``True``.
        scale_restoration (bool or str):
            Technique to restore scale ambiguity.
            If ``scale_restoration=True``, the projection back technique is applied to
            estimated spectrograms. You can also specify ``projection_back``
            or ``minimal_distortion_principle``. Default: ``True``.
        record_loss (bool):
            Record the loss at each iteration of the demixing filter update if ``record_loss=True``.
            Default: ``True``.
        reference_id (int):
            Reference channel for projection back and minimal distortion principle. Default: ``0``.

    Examples:
        Update demixing filters by IP:

        .. code-block:: python

            >>> def contrast_fn(y):
            ...     return 2 * np.abs(y)

            >>> def d_contrast_fn(y):
            ...     return 2 * np.ones_like(y)

            >>> n_channels, n_bins, n_frames = 2, 2049, 128
            >>> spectrogram_mix = np.random.randn(n_channels, n_bins, n_frames) \
            ...     + 1j * np.random.randn(n_channels, n_bins, n_frames)

            >>> fdica = AuxFDICA(
            ...     spatial_algorithm="IP",
            ...     contrast_fn=contrast_fn,
            ...     d_contrast_fn=d_contrast_fn,
            ... )
            >>> spectrogram_est = fdica(spectrogram_mix, n_iter=100)
            >>> print(spectrogram_mix.shape, spectrogram_est.shape)
            (2, 2049, 128), (2, 2049, 128)

        Update demixing filters by IP2:

        .. code-block:: python

            >>> from ssspy.utils.select_pair import sequential_pair_selector

            >>> def contrast_fn(y):
            ...     return 2 * np.abs(y)

            >>> def d_contrast_fn(y):
            ...     return 2 * np.ones_like(y)

            >>> n_channels, n_bins, n_frames = 2, 2049, 128
            >>> spectrogram_mix = np.random.randn(n_channels, n_bins, n_frames) \
            ...     + 1j * np.random.randn(n_channels, n_bins, n_frames)

            >>> fdica = AuxFDICA(
            ...     spatial_algorithm="IP2",
            ...     contrast_fn=contrast_fn,
            ...     d_contrast_fn=d_contrast_fn,
            ...     pair_selector=sequential_pair_selector,
            ... )
            >>> spectrogram_est = fdica(spectrogram_mix, n_iter=100)
            >>> print(spectrogram_mix.shape, spectrogram_est.shape)
            (2, 2049, 128), (2, 2049, 128)

    .. [#ono2010auxiliary]
        N. Ono and S. Miyabe,
        "Auxiliary-function-based independent component analysis for super-Gaussian sources,"
        in *Proc. LVA/ICA*, 2010, pp.165-172.
    """

    def __init__(
        self,
        spatial_algorithm: str = "IP",
        contrast_fn: Callable[[np.ndarray], np.ndarray] = None,
        d_contrast_fn: Callable[[np.ndarray], np.ndarray] = None,
        flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
            max_flooring, eps=EPS
        ),
        pair_selector: Optional[Callable[[int], Iterable[Tuple[int, int]]]] = None,
        callbacks: Optional[
            Union[Callable[["AuxFDICA"], None], List[Callable[["AuxFDICA"], None]]]
        ] = None,
        permutation_alignment: bool = True,
        scale_restoration: Union[bool, str] = True,
        record_loss: bool = True,
        reference_id: int = 0,
    ) -> None:
        super().__init__(
            contrast_fn=contrast_fn,
            flooring_fn=flooring_fn,
            callbacks=callbacks,
            permutation_alignment=permutation_alignment,
            scale_restoration=scale_restoration,
            record_loss=record_loss,
            reference_id=reference_id,
        )
        assert spatial_algorithm in spatial_algorithms, "Not support {}.".format(spatial_algorithms)

        self.spatial_algorithm = spatial_algorithm
        self.d_contrast_fn = d_contrast_fn

        if pair_selector is None:
            if spatial_algorithm == "IP2":
                self.pair_selector = sequential_pair_selector
        else:
            self.pair_selector = pair_selector

    def __call__(
        self, input: np.ndarray, n_iter: int = 100, initial_call: bool = True, **kwargs
    ) -> np.ndarray:
        r"""Separate a frequency-domain multichannel signal.

        Args:
            input (numpy.ndarray):
                The mixture signal in frequency-domain.
                The shape is (n_channels, n_bins, n_frames).
            n_iter (int):
                The number of iterations of demixing filter updates.
                Default: ``100``.
            initial_call (bool):
                If ``True``, perform callbacks (and computation of loss if necessary)
                before iterations.

        Returns:
            numpy.ndarray of the separated signal in frequency-domain.
            The shape is (n_channels, n_bins, n_frames).
        """
        self.input = input.copy()

        self._reset(**kwargs)

        # Call __call__ of FDICABase's parent, i.e. __call__ of IterativeMethodBase
        super(FDICABase, self).__call__(n_iter=n_iter, initial_call=initial_call)

        if self.permutation_alignment:
            self.solve_permutation()

        if self.scale_restoration:
            self.restore_scale()

        if self.demix_filter is not None:
            self.output = self.separate(self.input, demix_filter=self.demix_filter)
        else:
            # TODO: implement demixing-filter-free algorithms (e.g. ISS, IPA, etc.)
            pass

        return self.output

    def __repr__(self) -> str:
        s = "AuxFDICA("
        s += "spatial_algorithm={spatial_algorithm}"
        s += ", permutation_alignment={permutation_alignment}"
        s += ", scale_restoration={scale_restoration}"
        s += ", record_loss={record_loss}"

        if self.scale_restoration:
            s += ", reference_id={reference_id}"

        s += ")"

        return s.format(**self.__dict__)

    def update_once(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Update demixing filters once.

        - If ``self.spatial_algorithm`` is ``IP`` or ``IP1``, ``update_once_ip1`` is called.
        - If ``self.spatial_algorithm`` is ``IP2``, ``update_once_ip2`` is called.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.
        """
        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        if self.spatial_algorithm in ["IP", "IP1"]:
            self.update_once_ip1(flooring_fn=flooring_fn)
        elif self.spatial_algorithm in ["IP2"]:
            self.update_once_ip2(flooring_fn=flooring_fn)
        else:
            raise NotImplementedError("Not support {}.".format(self.spatial_algorithm))

    def update_once_ip1(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Update demixing filters once using iterative projection.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        Demixing filters are updated sequentially for :math:`n=1,\ldots,N` as follows:

        .. math::
            \boldsymbol{w}_{in}
            &\leftarrow\left(\boldsymbol{W}_{in}^{\mathsf{H}}\boldsymbol{U}_{in}\right)^{-1}
            \boldsymbol{e}_{n}, \\
            \boldsymbol{w}_{in}
            &\leftarrow\frac{\boldsymbol{w}_{in}}
            {\sqrt{\boldsymbol{w}_{in}^{\mathsf{H}}\boldsymbol{U}_{in}\boldsymbol{w}_{in}}}, \\

        where

        .. math::
            \boldsymbol{U}_{in}
            &= \frac{1}{J}\sum_{j}
            \frac{G'_{\mathbb{R}}(|y_{ijn}|)}{2|y_{ijn}|}
            \boldsymbol{x}_{ij}\boldsymbol{x}_{ij}^{\mathsf{H}}, \\
            G(y_{ijn})
            &= -\log p(y_{ijn}), \\
            G_{\mathbb{R}}(|y_{ijn}|)
            &= G(y_{ijn}).
        """
        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        X, W = self.input, self.demix_filter
        Y = self.separate(X, demix_filter=W)

        XX_Hermite = X[:, np.newaxis, :, :] * X[np.newaxis, :, :, :].conj()
        XX_Hermite = XX_Hermite.transpose(2, 0, 1, 3)  # (n_bins, n_channels, n_channels, n_frames)
        Y_abs = np.abs(Y)
        denom = flooring_fn(2 * Y_abs)
        varphi = self.d_contrast_fn(Y_abs) / denom  # (n_sources, n_bins, n_frames)
        varphi = varphi.transpose(1, 0, 2)  # (n_bins, n_sources, n_frames)
        GXX = varphi[:, :, np.newaxis, np.newaxis, :] * XX_Hermite[:, np.newaxis, :, :, :]
        U = np.mean(GXX, axis=-1)  # (n_bins, n_sources, n_channels, n_channels)

        self.demix_filter = update_by_ip1(W, U, flooring_fn=flooring_fn)

    def update_once_ip2(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Update demixing filters once using pairwise iterative projection.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        For :math:`n_{1}` and :math:`n_{2}` (:math:`n_{1}\neq n_{2}`),
        compute auxiliary variables:

        .. math::
            \bar{r}_{ijn_{1}}
            &\leftarrow|y_{ijn_{1}}| \\
            \bar{r}_{ijn_{2}}
            &\leftarrow|y_{ijn_{2}}|

        Then, for :math:`n=n_{1},n_{2}`, compute weighted covariance matrix as follows:

        .. math::
            \boldsymbol{U}_{in_{1}}
            &= \frac{1}{J}\sum_{j}
            \frac{G'_{\mathbb{R}}(\bar{r}_{ijn_{1}})}{2\bar{r}_{ijn_{1}}}
            \boldsymbol{x}_{ij}\boldsymbol{x}_{ij}^{\mathsf{H}}, \\
            \boldsymbol{U}_{in_{2}}
            &= \frac{1}{J}\sum_{j}
            \frac{G'_{\mathbb{R}}(\bar{r}_{ijn_{2}})}{2\bar{r}_{ijn_{2}}}
            \boldsymbol{x}_{ij}\boldsymbol{x}_{ij}^{\mathsf{H}},

        where

        .. math::
            G(y_{ijn})
            &= -\log p(y_{ijn}), \\
            G_{\mathbb{R}}(|y_{ijn}|)
            &= G(y_{ijn}).

        Using :math:`\boldsymbol{U}_{in_{1}}` and
        :math:`\boldsymbol{U}_{in_{2}}`, we compute generalized eigenvectors.

        .. math::
            \left({\boldsymbol{P}_{in_{1}}^{(n_{1},n_{2})}}^{\mathsf{H}}\boldsymbol{U}_{in_{1}}
            \boldsymbol{P}_{in_{1}}^{(n_{1},n_{2})}\right)\boldsymbol{h}_{i}
            = \lambda_{i}
            \left({\boldsymbol{P}_{in_{2}}^{(n_{1},n_{2})}}^{\mathsf{H}}\boldsymbol{U}_{in_{2}}
            \boldsymbol{P}_{in_{2}}^{(n_{1},n_{2})}\right)\boldsymbol{h}_{i},

        where

        .. math::
            \boldsymbol{P}_{in_{1}}^{(n_{1},n_{2})}
            &= (\boldsymbol{W}_{i}\boldsymbol{U}_{in_{1}})^{-1}
            (
            \begin{array}{cc}
                \boldsymbol{e}_{n_{1}} & \boldsymbol{e}_{n_{2}}
            \end{array}
            ), \\
            \boldsymbol{P}_{in_{2}}^{(n_{1},n_{2})}
            &= (\boldsymbol{W}_{i}\boldsymbol{U}_{in_{2}})^{-1}
            (
            \begin{array}{cc}
                \boldsymbol{e}_{n_{1}} & \boldsymbol{e}_{n_{2}}
            \end{array}
            ).

        After that, we standardize two eigenvectors :math:`\boldsymbol{h}_{in_{1}}`
        and :math:`\boldsymbol{h}_{in_{2}}`.

        .. math::
            \boldsymbol{h}_{in_{1}}
            &\leftarrow\frac{\boldsymbol{h}_{in_{1}}}
            {\sqrt{\boldsymbol{h}_{in_{1}}^{\mathsf{H}}
            \left({\boldsymbol{P}_{in_{1}}^{(n_{1},n_{2})}}^{\mathsf{H}}\boldsymbol{U}_{in_{1}}
            \boldsymbol{P}_{in_{1}}^{(n_{1},n_{2})}\right)
            \boldsymbol{h}_{in_{1}}}}, \\
            \boldsymbol{h}_{in_{2}}
            &\leftarrow\frac{\boldsymbol{h}_{in_{2}}}
            {\sqrt{\boldsymbol{h}_{in_{2}}^{\mathsf{H}}
            \left({\boldsymbol{P}_{in_{2}}^{(n_{1},n_{2})}}^{\mathsf{H}}\boldsymbol{U}_{in_{2}}
            \boldsymbol{P}_{in_{2}}^{(n_{1},n_{2})}\right)
            \boldsymbol{h}_{in_{2}}}}.

        Then, update :math:`\boldsymbol{w}_{in_{1}}` and :math:`\boldsymbol{w}_{in_{2}}`
        simultaneously.

        .. math::
            \boldsymbol{w}_{in_{1}}
            &\leftarrow \boldsymbol{P}_{in_{1}}^{(n_{1},n_{2})}\boldsymbol{h}_{in_{1}} \\
            \boldsymbol{w}_{in_{2}}
            &\leftarrow \boldsymbol{P}_{in_{2}}^{(n_{1},n_{2})}\boldsymbol{h}_{in_{2}}

        At each iteration, we update pairs of :math:`n_{1}` and :math:`n_{1}`
        for :math:`n_{1}\neq n_{2}`.
        """
        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        n_sources = self.n_sources
        X, W = self.input, self.demix_filter

        XX_Hermite = X[:, np.newaxis, :, :] * X[np.newaxis, :, :, :].conj()
        XX_Hermite = XX_Hermite.transpose(2, 0, 1, 3)

        for m, n in self.pair_selector(n_sources):
            W_mn = W[:, (m, n), :]
            Y_mn = self.separate(X, demix_filter=W_mn)

            Y_abs_mn = np.abs(Y_mn)
            denom = flooring_fn(2 * Y_abs_mn)
            varphi_mn = self.d_contrast_fn(Y_abs_mn) / denom
            varphi_mn = varphi_mn.transpose(1, 0, 2)
            GXX_mn = varphi_mn[:, :, np.newaxis, np.newaxis, :] * XX_Hermite[:, np.newaxis, :, :, :]
            U_mn = np.mean(GXX_mn, axis=-1)

            W[:, (m, n), :] = update_by_ip2_one_pair(
                W,
                U_mn,
                pair=(m, n),
                flooring_fn=flooring_fn,
            )

        self.demix_filter = W


class GradLaplaceFDICA(GradFDICA):
    r"""Frequency-domain independent component analysis (FDICA) \
    using the gradient descent on a Laplace distribution.

    We assume :math:`y_{ijn}` follows a Laplace distribution.

    .. math::
        p(y_{ijn})\propto\exp(|y_{ijn}|)

    Args:
        step_size (float):
            A step size of the gradient descent. Default: ``1e-1``.
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to receive (n_channels, n_bins, n_frames)
            and return (n_channels, n_bins, n_frames).
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``partial(max_flooring, eps=1e-10)``.
        callbacks (callable or list[callable], optional):
            Callback functions. Each function is called before separation and at each iteration.
            Default: ``None``.
        is_holonomic (bool):
            If ``is_holonomic=True``, Holonomic-type update is used.
            Otherwise, Nonholonomic-type update is used. Default: ``False``.
        permutation_alignment (bool):
            If ``permutation_alignment=True``, a permutation solver is used to align
            estimated spectrograms. Default: ``True``.
        scale_restoration (bool or str):
            Technique to restore scale ambiguity.
            If ``scale_restoration=True``, the projection back technique is applied to
            estimated spectrograms. You can also specify ``projection_back``
            or ``minimal_distortion_principle``. Default: ``True``.
        record_loss (bool):
            Record the loss at each iteration of the gradient descent if ``record_loss=True``.
            Default: ``True``.
        reference_id (int):
            Reference channel for projection back and minimal distortion principle. Default: ``0``.

    Examples:
        Update demixing filters using Holonomic-type update:

        .. code-block:: python

            >>> n_channels, n_bins, n_frames = 2, 2049, 128
            >>> spectrogram_mix = \
            ...     np.random.randn(n_channels, n_bins, n_frames) \
            ...     + 1j * np.random.randn(n_channels, n_bins, n_frames)

            >>> fdica = GradLaplaceFDICA(is_holonomic=True)
            >>> spectrogram_est = fdica(spectrogram_mix, n_iter=1000)
            >>> print(spectrogram_mix.shape, spectrogram_est.shape)
            (2, 2049, 128), (2, 2049, 128)

        Update demixing filters using Nonholonomic-type update:

        .. code-block:: python

            >>> n_channels, n_bins, n_frames = 2, 2049, 128
            >>> spectrogram_mix = \
            ...     np.random.randn(n_channels, n_bins, n_frames) \
            ...     + 1j * np.random.randn(n_channels, n_bins, n_frames)

            >>> fdica = GradLaplaceFDICA(is_holonomic=False)
            >>> spectrogram_est = fdica(spectrogram_mix, n_iter=1000)
            >>> print(spectrogram_mix.shape, spectrogram_est.shape)
            (2, 2049, 128), (2, 2049, 128)
    """

    def __init__(
        self,
        step_size: float = 1e-1,
        flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
            max_flooring, eps=EPS
        ),
        callbacks: Optional[
            Union[Callable[["GradLaplaceFDICA"], None], List[Callable[["GradLaplaceFDICA"], None]]]
        ] = None,
        is_holonomic: bool = False,
        permutation_alignment: bool = True,
        scale_restoration: Union[bool, str] = True,
        record_loss: bool = True,
        reference_id: int = 0,
    ) -> None:
        def contrast_fn(y: np.ndarray) -> np.ndarray:
            r"""Contrast function.

            Args:
                y (numpy.ndarray):
                    The shape is (n_sources, n_bins, n_frames).

            Returns:
                The shape is (n_sources, n_bins, n_frames).
            """
            return 2 * np.abs(y)

        def score_fn(y: np.ndarray) -> np.ndarray:
            r"""Score function.

            Args:
                y (numpy.ndarray):
                    The shape is (n_sources, n_bins, n_frames).

            Returns:
                The shape is (n_sources, n_bins, n_frames).
            """
            denom = self.flooring_fn(np.abs(y))
            return y / denom

        super().__init__(
            step_size=step_size,
            contrast_fn=contrast_fn,
            score_fn=score_fn,
            flooring_fn=flooring_fn,
            callbacks=callbacks,
            is_holonomic=is_holonomic,
            permutation_alignment=permutation_alignment,
            scale_restoration=scale_restoration,
            record_loss=record_loss,
            reference_id=reference_id,
        )

    def __repr__(self) -> str:
        s = "GradLaplaceFDICA("
        s += "step_size={step_size}"
        s += ", is_holonomic={is_holonomic}"
        s += ", permutation_alignment={permutation_alignment}"
        s += ", scale_restoration={scale_restoration}"
        s += ", record_loss={record_loss}"

        if self.scale_restoration:
            s += ", reference_id={reference_id}"

        s += ")"

        return s.format(**self.__dict__)


class NaturalGradLaplaceFDICA(NaturalGradFDICA):
    r"""Frequency-domain independent component analysis (FDICA) \
    using the natural gradient descent on a Laplace distribution.

    We assume :math:`y_{ijn}` follows a Laplace distribution.

    .. math::
        p(y_{ijn})\propto\exp(|y_{ijn}|)

    Args:
        step_size (float):
            A step size of the gradient descent. Default: ``1e-1``.
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to receive (n_channels, n_bins, n_frames)
            and return (n_channels, n_bins, n_frames).
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``partial(max_flooring, eps=1e-10)``.
        callbacks (callable or list[callable], optional):
            Callback functions. Each function is called before separation and at each iteration.
            Default: ``None``.
        is_holonomic (bool):
            If ``is_holonomic=True``, Holonomic-type update is used.
            Otherwise, Nonholonomic-type update is used. Default: ``False``.
        permutation_alignment (bool):
            If ``permutation_alignment=True``, a permutation solver is used to align
            estimated spectrograms. Default: ``True``.
        scale_restoration (bool or str):
            Technique to restore scale ambiguity.
            If ``scale_restoration=True``, the projection back technique is applied to
            estimated spectrograms. You can also specify ``projection_back``
            or ``minimal_distortion_principle``. Default: ``True``.
        record_loss (bool):
            Record the loss at each iteration of the gradient descent if ``record_loss=True``.
            Default: ``True``.
        reference_id (int):
            Reference channel for projection back and minimal distortion principle. Default: ``0``.

    Examples:
        Update demixing filters using Holonomic-type update:

        .. code-block:: python

            >>> n_channels, n_bins, n_frames = 2, 2049, 128
            >>> spectrogram_mix = \
            ...     np.random.randn(n_channels, n_bins, n_frames) \
            ...     + 1j * np.random.randn(n_channels, n_bins, n_frames)

            >>> fdica = NaturalGradLaplaceFDICA(is_holonomic=True)
            >>> spectrogram_est = fdica(spectrogram_mix, n_iter=1000)
            >>> print(spectrogram_mix.shape, spectrogram_est.shape)
            (2, 2049, 128), (2, 2049, 128)

        Update demixing filters using Nonholonomic-type update:

        .. code-block:: python

            >>> n_channels, n_bins, n_frames = 2, 2049, 128
            >>> spectrogram_mix = \
            ...     np.random.randn(n_channels, n_bins, n_frames) \
            ...     + 1j * np.random.randn(n_channels, n_bins, n_frames)

            >>> fdica = NaturalGradLaplaceFDICA(is_holonomic=False)
            >>> spectrogram_est = fdica(spectrogram_mix, n_iter=1000)
            >>> print(spectrogram_mix.shape, spectrogram_est.shape)
            (2, 2049, 128), (2, 2049, 128)
    """

    def __init__(
        self,
        step_size: float = 1e-1,
        flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
            max_flooring, eps=EPS
        ),
        callbacks: Optional[
            Union[
                Callable[["NaturalGradLaplaceFDICA"], None],
                List[Callable[["NaturalGradLaplaceFDICA"], None]],
            ]
        ] = None,
        is_holonomic: bool = False,
        permutation_alignment: bool = True,
        scale_restoration: Union[bool, str] = True,
        record_loss: bool = True,
        reference_id: int = 0,
    ) -> None:
        def contrast_fn(y: np.ndarray) -> np.ndarray:
            r"""Contrast function.

            Args:
                y (numpy.ndarray):
                    The shape is (n_sources, n_bins, n_frames).

            Returns:
                The shape is (n_sources, n_bins, n_frames).
            """
            return 2 * np.abs(y)

        def score_fn(y: np.ndarray) -> np.ndarray:
            r"""Score function.

            Args:
                y (numpy.ndarray):
                    The shape is (n_sources, n_bins, n_frames).

            Returns:
                The shape is (n_sources, n_bins, n_frames).
            """
            denom = self.flooring_fn(np.abs(y))
            return y / denom

        super().__init__(
            step_size=step_size,
            contrast_fn=contrast_fn,
            score_fn=score_fn,
            flooring_fn=flooring_fn,
            callbacks=callbacks,
            is_holonomic=is_holonomic,
            permutation_alignment=permutation_alignment,
            scale_restoration=scale_restoration,
            record_loss=record_loss,
            reference_id=reference_id,
        )

    def __repr__(self) -> str:
        s = "NaturalGradLaplaceFDICA("
        s += "step_size={step_size}"
        s += ", is_holonomic={is_holonomic}"
        s += ", permutation_alignment={permutation_alignment}"
        s += ", scale_restoration={scale_restoration}"
        s += ", record_loss={record_loss}"

        if self.scale_restoration:
            s += ", reference_id={reference_id}"

        s += ")"

        return s.format(**self.__dict__)


class AuxLaplaceFDICA(AuxFDICA):
    r"""Auxiliary-function-based frequency-domain independent component analysis \
    on a Laplace distribution.

    We assume :math:`y_{ijn}` follows a Laplace distribution.

    .. math::
        p(y_{ijn})\propto\exp(|y_{ijn}|)

    Args:
        spatial_algorithm (str):
            Algorithm to update demixing filters.
            Choose ``IP``, ``IP1``, or ``IP2``. Default: ``IP``.
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to receive (n_channels, n_bins, n_frames)
            and return (n_channels, n_bins, n_frames).
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``partial(max_flooring, eps=1e-10)``.
        pair_selector (callable, optional):
            Selector to choose updaing pair in ``IP2`` and ``ISS2``.
            If ``None`` is given, ``partial(sequential_pair_selector, sort=True)`` is used.
            Default: ``None``.
        callbacks (callable or list[callable], optional):
            Callback functions. Each function is called before separation and at each iteration.
            Default: ``None``.
        permutation_alignment (bool):
            If ``permutation_alignment=True``, a permutation solver is used to align
            estimated spectrograms. Default: ``True``.
        scale_restoration (bool or str):
            Technique to restore scale ambiguity.
            If ``scale_restoration=True``, the projection back technique is applied to
            estimated spectrograms. You can also specify ``projection_back``
            or ``minimal_distortion_principle``. Default: ``True``.
        record_loss (bool):
            Record the loss at each iteration of the demixing filter update if ``record_loss=True``.
            Default: ``True``.
        reference_id (int):
            Reference channel for projection back and minimal distortion principle. Default: ``0``.

    Examples:
        Update demixing filters by IP:

        .. code-block:: python

            >>> n_channels, n_bins, n_frames = 2, 2049, 128
            >>> spectrogram_mix = \
            ...     np.random.randn(n_channels, n_bins, n_frames) \
            ...     + 1j * np.random.randn(n_channels, n_bins, n_frames)

            >>> fdica = AuxLaplaceFDICA(spatial_algorithm="IP")
            >>> spectrogram_est = fdica(spectrogram_mix, n_iter=1000)
            >>> print(spectrogram_mix.shape, spectrogram_est.shape)
            (2, 2049, 128), (2, 2049, 128)

        Update demixing filters by IP2:

        .. code-block:: python

            >>> from ssspy.utils.select_pair import sequential_pair_selector

            >>> n_channels, n_bins, n_frames = 2, 2049, 128
            >>> spectrogram_mix = \
            ...     np.random.randn(n_channels, n_bins, n_frames) \
            ...     + 1j * np.random.randn(n_channels, n_bins, n_frames)

            >>> fdica = AuxLaplaceFDICA(
            ...     spatial_algorithm="IP2",
            ...     pair_selector=sequential_pair_selector,
            ... )
            >>> spectrogram_est = fdica(spectrogram_mix, n_iter=1000)
            >>> print(spectrogram_mix.shape, spectrogram_est.shape)
            (2, 2049, 128), (2, 2049, 128)
    """

    def __init__(
        self,
        spatial_algorithm: str = "IP",
        flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
            max_flooring, eps=EPS
        ),
        pair_selector: Optional[Callable[[int], Iterable[Tuple[int, int]]]] = None,
        callbacks: Optional[
            Union[Callable[["AuxLaplaceFDICA"], None], List[Callable[["AuxLaplaceFDICA"], None]]]
        ] = None,
        permutation_alignment: bool = True,
        scale_restoration: Union[bool, str] = True,
        record_loss: bool = True,
        reference_id: int = 0,
    ) -> None:
        def contrast_fn(y: np.ndarray):
            r"""Contrast function.

            Args:
                y (numpy.ndarray):
                    The shape is (n_sources, n_bins, n_frames).

            Returns:
                The shape is (n_sources, n_bins, n_frames).
            """
            return 2 * np.abs(y)

        def d_contrast_fn(y: np.ndarray):
            r"""Partial derivative of score function.

            Args:
                y (numpy.ndarray):
                    The shape is (n_sources, n_bins, n_frames).

            Returns:
                The shape is (n_sources, n_bins, n_frames).
            """
            return 2 * np.ones_like(y)

        super().__init__(
            spatial_algorithm=spatial_algorithm,
            contrast_fn=contrast_fn,
            d_contrast_fn=d_contrast_fn,
            flooring_fn=flooring_fn,
            pair_selector=pair_selector,
            callbacks=callbacks,
            permutation_alignment=permutation_alignment,
            scale_restoration=scale_restoration,
            record_loss=record_loss,
            reference_id=reference_id,
        )

    def __repr__(self) -> str:
        s = "AuxLaplaceFDICA("
        s += "spatial_algorithm={spatial_algorithm}"
        s += ", permutation_alignment={permutation_alignment}"
        s += ", scale_restoration={scale_restoration}"
        s += ", record_loss={record_loss}"

        if self.scale_restoration:
            s += ", reference_id={reference_id}"

        s += ")"

        return s.format(**self.__dict__)



import os
import shutil
import urllib.request

import numpy as np

reverb_durations = [0.16, 0.36, 0.61]


def download(root: str = ".data/MIRD", n_sources: int = 3, reverb_duration: float = 0.16) -> str:
    assert reverb_duration in reverb_durations, "reverb_duration should be chosen from {}.".format(
        reverb_durations
    )

    filename = (
        "Impulse_response_Acoustic_Lab_Bar-Ilan_University__"
        "Reverberation_{reverb_duration:.3f}s__3-3-3-8-3-3-3.zip"
    )
    filename = filename.format(reverb_duration=reverb_duration)
    url = (
        "https://www.iks.rwth-aachen.de/fileadmin/user_upload/downloads/"
        "forschung/tools-downloads/{filename}"
    )
    url = url.format(filename=filename)
    zip_path = os.path.join(root, filename)

    degrees = [30, 345, 0, 60, 315]
    channels = [3, 4, 2, 5, 1, 6, 0, 7]
    sample_rate = 16000
    duration = reverb_duration

    degrees = degrees[:n_sources]
    channels = channels[:n_sources]

    n_channels = len(channels)
    n_samples = int(sample_rate * duration)

    template_rir_name = (
        "Impulse_response_Acoustic_Lab_Bar-Ilan_University_"
        "(Reverberation_{:.3f}s)_3-3-3-8-3-3-3_1m_{:03d}.mat"
    )

    os.makedirs(root, exist_ok=True)

    if not os.path.exists(zip_path):
        urllib.request.urlretrieve(url, zip_path)

    rir_path = os.path.join(root, template_rir_name.format(reverb_duration, 0))

    if not os.path.exists(rir_path):
        shutil.unpack_archive(zip_path, root)

    npz_path = os.path.join(root, "MIRD-{}ch.npz".format(n_channels))

    assert n_channels == n_sources, "Mixing system should be determined."

    if not os.path.exists(npz_path):
        rirs = {}

        for src_idx, degree in enumerate(degrees):
            rir_path = os.path.join(root, template_rir_name.format(duration, degree))
            rir = resample_mird_rir(rir_path, sample_rate_out=sample_rate)
            rirs["src_{}".format(src_idx + 1)] = rir[channels, :n_samples]

        np.savez(
            npz_path, sample_rate=sample_rate, n_sources=n_sources, n_channels=n_channels, **rirs
        )

    return npz_path


def resample_mird_rir(rir_path: str, sample_rate_out: int) -> np.ndarray:
    import scipy.signal as ss
    from scipy.io import loadmat

    sample_rate_in = 48000
    rir_mat = loadmat(rir_path)
    rir = rir_mat["impulse_response"]

    rir_resampled = ss.resample_poly(rir, sample_rate_out, sample_rate_in, axis=0)

    return rir_resampled.T



import functools
from typing import Callable, List, Optional, Union

import numpy as np

from ..algorithm.permutation_alignment import (
    correlation_based_permutation_solver,
    score_based_permutation_solver,
)
from ..linalg.quadratic import quadratic
from ..special.flooring import identity, max_flooring
from ..special.logsumexp import logsumexp
from ..special.psd import to_psd
from ..special.softmax import softmax
from ..utils.flooring import choose_flooring_fn
from .base import IterativeMethodBase

EPS = 1e-10


class CACGMMBase(IterativeMethodBase):
    r"""Base class of complex angular central Gaussian mixture model (cACGMM).

    Args:
        n_sources (int, optional):
            Number of sources to be separated.
            If ``None`` is given, ``n_sources`` is determined by number of channels
            in input spectrogram. Default: ``None``.
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to return the same shape tensor as the input.
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
        callbacks (callable or list[callable], optional):
            Callback functions. Each function is called before separation and at each iteration.
            Default: ``None``.
        record_loss (bool):
            Record the loss at each iteration of the update algorithm if ``record_loss=True``.
            Default: ``True``.
        rng (numpy.random.Generator, optioinal):
            Random number generator. This is mainly used to randomly initialize parameters
            of cACGMM. If ``None`` is given, ``np.random.default_rng()`` is used.
            Default: ``None``.
    """

    def __init__(
        self,
        n_sources: Optional[int] = None,
        flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
            max_flooring, eps=EPS
        ),
        callbacks: Optional[
            Union[
                Callable[["CACGMMBase"], None],
                List[Callable[["CACGMMBase"], None]],
            ]
        ] = None,
        record_loss: bool = True,
        rng: Optional[np.random.Generator] = None,
    ) -> None:
        self.normalization: bool
        self.permutation_alignment: bool

        super().__init__(callbacks=callbacks, record_loss=record_loss)

        self.n_sources = n_sources

        if flooring_fn is None:
            self.flooring_fn = identity
        else:
            self.flooring_fn = flooring_fn

        if rng is None:
            rng = np.random.default_rng()

        self.rng = rng

    def __call__(
        self, input: np.ndarray, n_iter: int = 100, initial_call: bool = True, **kwargs
    ) -> np.ndarray:
        r"""Separate a frequency-domain multichannel signal.

        Args:
            input (numpy.ndarray):
                The mixture signal in frequency-domain.
                The shape is (n_channels, n_bins, n_frames).
            n_iter (int):
                The number of iterations of demixing filter updates.
                Default: ``100``.
            initial_call (bool):
                If ``True``, perform callbacks (and computation of loss if necessary)
                before iterations.

        Returns:
            numpy.ndarray of the separated signal in frequency-domain.
            The shape is (n_channels, n_bins, n_frames).
        """
        self.input = input.copy()

        self._reset(**kwargs)

        raise NotImplementedError("Implement '__call__' method.")

    def __repr__(self) -> str:
        s = "CACGMM("

        if self.n_sources is not None:
            s += "n_sources={n_sources}, "

        s += "record_loss={record_loss}"

        s += ")"

        return s.format(**self.__dict__)

    def _reset(self, **kwargs) -> None:
        r"""Reset attributes by given keyword arguments.

        Args:
            kwargs:
                Keyword arguments to set as attributes of CACGMM.
        """
        assert self.input is not None, "Specify data!"

        for key in kwargs.keys():
            setattr(self, key, kwargs[key])

        X = self.input

        norm = np.linalg.norm(X, axis=0)
        Z = X / norm
        self.unit_input = Z

        n_sources = self.n_sources
        n_channels, n_bins, n_frames = X.shape

        if n_sources is None:
            n_sources = n_channels

        self.n_sources, self.n_channels = n_sources, n_channels
        self.n_bins, self.n_frames = n_bins, n_frames

        self._init_parameters(rng=self.rng)

    def _init_parameters(self, rng: Optional[np.random.Generator] = None) -> None:
        r"""Initialize parameters of cACGMM.

        Args:
            rng (numpy.random.Generator, optional):
                Random number generator. If ``None`` is given,
                ``np.random.default_rng()`` is used.
                Default: ``None``.
        """
        n_sources, n_channels = self.n_sources, self.n_channels
        n_bins = self.n_bins

        if rng is None:
            rng = np.random.default_rng()

        alpha = rng.random((n_sources, n_bins))
        alpha = alpha / alpha.sum(axis=0)

        eye = np.eye(n_channels, dtype=np.complex128)
        B_diag = self.rng.random((n_sources, n_bins, n_channels))
        B_diag = B_diag / B_diag.sum(axis=-1, keepdims=True)
        B = B_diag[:, :, :, np.newaxis] * eye

        self.mixing = alpha
        self.covariance = B

        # The shape of posterior is (n_sources, n_bins, n_frames).
        # This is always required to satisfy posterior.sum(axis=0) = 1
        self.posterior = None

    def separate(self, input: np.ndarray) -> np.ndarray:
        r"""Separate ``input``.

        Args:
            input (numpy.ndarray):
                The mixture signal in frequency-domain.
                The shape is (n_channels, n_bins, n_frames).

        Returns:
            numpy.ndarray of the separated signal in frequency-domain.
            The shape is (n_sources, n_bins, n_frames).
        """
        raise NotImplementedError("Implement 'separate' method.")

    def normalize_covariance(self) -> None:
        r"""Normalize covariance of cACG.

        .. math::
            \boldsymbol{B}_{in}
            \leftarrow\frac{\boldsymbol{B}_{in}}{\mathrm{tr}(\boldsymbol{B}_{in})}
        """
        assert self.normalization, "Set normalization."

        B = self.covariance

        trace = np.trace(B, axis1=-2, axis2=-1)
        trace = np.real(trace)
        B = B / trace[..., np.newaxis, np.newaxis]

        self.covariance = B

    def compute_loss(self) -> float:
        r"""Compute loss :math:`\mathcal{L}`.

        Returns:
            Computed loss.
        """
        raise NotImplementedError("Implement 'compute_loss' method.")

    def compute_logdet(self, covariance: np.ndarray) -> np.ndarray:
        r"""Compute log-determinant of input.

        Args:
            covariance (numpy.ndarray):
                Covariance matrix with shape of (n_sources, n_bins, n_channels, n_channels).

        Returns:
            numpy.ndarray of log-determinant.
        """
        _, logdet = np.linalg.slogdet(covariance)

        return logdet

    def solve_permutation(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Align posteriors and separated spectrograms.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        """

        permutation_alignment = self.permutation_alignment
        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        assert permutation_alignment, "Set permutation_alignment=True."

        if type(permutation_alignment) is bool:
            # when permutation_alignment is True
            permutation_alignment = "posterior_score"

        if permutation_alignment in ["posterior_score", "posterior_correlation"]:
            target = "posterior"
        elif permutation_alignment in ["amplitude_score", "amplitude_correlation"]:
            target = "amplitude"
        else:
            raise NotImplementedError(
                "permutation_alignment {} is not implemented.".format(permutation_alignment)
            )

        if permutation_alignment in ["posterior_score", "amplitude_score"]:
            self.solve_permutation_by_score(target=target, flooring_fn=flooring_fn)
        elif permutation_alignment in ["posterior_correlation", "amplitude_correlation"]:
            self.solve_permutation_by_correlation(target=target, flooring_fn=flooring_fn)
        else:
            raise NotImplementedError(
                "permutation_alignment {} is not implemented.".format(permutation_alignment)
            )

    def solve_permutation_by_score(
        self,
        target: str = "posterior",
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Align posteriors and amplitudes of separated spectrograms by score value.

        Args:
            target (str):
                Target to compute score values. Choose ``posterior`` or ``amplitude``.
                Default: ``posterior``.
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.
        """

        assert target in ["posterior", "amplitude"], "Invalid target {} is specified.".format(
            target
        )

        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        X = self.input
        alpha = self.mixing
        B = self.covariance
        gamma = self.posterior

        if hasattr(self, "global_iter"):
            global_iter = self.global_iter
        else:
            global_iter = 1

        if hasattr(self, "local_iter"):
            local_iter = self.local_iter
        else:
            local_iter = 1

        Y = self.separate(X, posterior=gamma)

        alpha = alpha.transpose(1, 0)
        B = B.transpose(1, 0, 2, 3)
        gamma = gamma.transpose(1, 0, 2)

        if target == "posterior":
            gamma, (alpha, B) = score_based_permutation_solver(
                gamma,
                alpha,
                B,
                global_iter=global_iter,
                local_iter=local_iter,
                flooring_fn=flooring_fn,
            )
        elif target == "amplitude":
            Y = Y.transpose(1, 0, 2)
            amplitude = np.abs(Y)

            _, (alpha, B, gamma) = score_based_permutation_solver(
                amplitude,
                alpha,
                B,
                gamma,
                global_iter=global_iter,
                local_iter=local_iter,
                flooring_fn=flooring_fn,
            )
        else:
            raise ValueError("Invalid target {} is specified.".format(target))

        alpha = alpha.transpose(1, 0)
        B = B.transpose(1, 0, 2, 3)
        gamma = gamma.transpose(1, 0, 2)

        Y = self.separate(X, posterior=gamma)

        self.mixing = alpha
        self.covariance = B
        self.posterior = gamma
        self.output = Y

    def solve_permutation_by_correlation(
        self,
        target: str = "amplitude",
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Align posteriors and amplitudes of separated spectrograms by correlation.

        Args:
            target (str):
                Target to compute correlations. Choose ``posterior`` or ``amplitude``.
                Default: ``amplitude``.
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        """

        assert target == "amplitude", "Only amplitude is supported as target."

        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        X = self.input
        alpha = self.mixing
        B = self.covariance
        gamma = self.posterior

        Y = self.separate(X, posterior=self.posterior)

        alpha = alpha.transpose(1, 0)
        B = B.transpose(1, 0, 2, 3)
        gamma = gamma.transpose(1, 0, 2)
        Y = Y.transpose(1, 0, 2)
        Y, (alpha, B, gamma) = correlation_based_permutation_solver(
            Y, alpha, B, gamma, flooring_fn=flooring_fn
        )
        alpha = alpha.transpose(1, 0)
        B = B.transpose(1, 0, 2, 3)
        gamma = gamma.transpose(1, 0, 2)
        Y = Y.transpose(1, 0, 2)

        self.mixing = alpha
        self.covariance = B
        self.posterior = gamma
        self.output = Y


class CACGMM(CACGMMBase):
    r"""Complex angular central Gaussian mixture model (cACGMM) [#ito2016complex]_.

    Args:
        n_sources (int, optional):
            Number of sources to be separated.
            If ``None`` is given, ``n_sources`` is determined by number of channels
            in input spectrogram. Default: ``None``.
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to return the same shape tensor as the input.
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
        callbacks (callable or list[callable], optional):
            Callback functions. Each function is called before separation and at each iteration.
            Default: ``None``.
        normalization (bool):
            If ``True`` is given, normalization is applied to covariance in cACG.
        permutation_alignment (bool):
            If ``permutation_alignment=True``, a permutation solver is used to align
            estimated spectrograms. Default: ``True``.
        record_loss (bool):
            Record the loss at each iteration of the update algorithm if ``record_loss=True``.
            Default: ``True``.
        reference_id (int):
            Reference channel to extract separated signals. Default: ``0``.
        rng (numpy.random.Generator, optioinal):
            Random number generator. This is mainly used to randomly initialize parameters
            of cACGMM. If ``None`` is given, ``np.random.default_rng()`` is used.
            Default: ``None``.

    .. [#ito2016complex] N. Ito, S. Araki, and T. Nakatani. \
        "Complex angular central Gaussian mixture model for directional statistics \
        in mask-based microphone array signal processing,"
        in *Proc. EUSIPCO*, 2016, pp. 1153-1157.
    """

    def __init__(
        self,
        n_sources: Optional[int] = None,
        flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
            max_flooring, eps=EPS
        ),
        callbacks: Optional[
            Union[
                Callable[["CACGMM"], None],
                List[Callable[["CACGMM"], None]],
            ]
        ] = None,
        normalization: bool = True,
        permutation_alignment: bool = True,
        record_loss: bool = True,
        reference_id: int = 0,
        rng: Optional[np.random.Generator] = None,
        **kwargs,
    ) -> None:
        super().__init__(
            n_sources=n_sources,
            flooring_fn=flooring_fn,
            callbacks=callbacks,
            record_loss=record_loss,
            rng=rng,
        )

        self.normalization = normalization
        self.permutation_alignment = permutation_alignment
        self.reference_id = reference_id

        if type(permutation_alignment) is bool and permutation_alignment:
            valid_keys = {"global_iter", "local_iter"}
        elif type(permutation_alignment) is str and permutation_alignment in [
            "posterior_score",
            "amplitude_score",
        ]:
            valid_keys = {"global_iter", "local_iter"}
        else:
            valid_keys = set()

        invalid_keys = set(kwargs) - valid_keys

        assert invalid_keys == set(), "Invalid keywords {} are given.".format(invalid_keys)

        for key, value in kwargs.items():
            setattr(self, key, value)

    def __call__(
        self, input: np.ndarray, n_iter: int = 100, initial_call: bool = True, **kwargs
    ) -> np.ndarray:
        r"""Separate a frequency-domain multichannel signal.

        Args:
            input (numpy.ndarray):
                The mixture signal in frequency-domain.
                The shape is (n_channels, n_bins, n_frames).
            n_iter (int):
                The number of iterations of demixing filter updates.
                Default: ``100``.
            initial_call (bool):
                If ``True``, perform callbacks (and computation of loss if necessary)
                before iterations.

        Returns:
            numpy.ndarray of the separated signal in frequency-domain.
            The shape is (n_channels, n_bins, n_frames).
        """
        self.input = input.copy()

        self._reset(**kwargs)

        # Call __call__ of CACGMMBase's parent, i.e. __call__ of IterativeMethodBase
        super(CACGMMBase, self).__call__(n_iter=n_iter, initial_call=initial_call)

        # posterior should be updated
        self.update_posterior(flooring_fn=self.flooring_fn)

        if self.permutation_alignment:
            self.solve_permutation(flooring_fn=self.flooring_fn)

        X = self.input
        self.output = self.separate(X, posterior=self.posterior)

        return self.output

    def __repr__(self) -> str:
        s = "CACGMM("

        if self.n_sources is not None:
            s += "n_sources={n_sources}, "

        s += "record_loss={record_loss}"
        s += ", normalization={normalization}"
        s += ", permutation_alignment={permutation_alignment}"
        s += ", reference_id={reference_id}"

        s += ")"

        return s.format(**self.__dict__)

    def separate(self, input: np.ndarray, posterior: Optional[np.ndarray] = None) -> np.ndarray:
        r"""Separate ``input`` using posterior probabilities.

        In this method, ``self.posterior`` is not updated.

        Args:
            input (numpy.ndarray):
                The mixture signal in frequency-domain.
                The shape is (n_channels, n_bins, n_frames).
            posterior (numpy.ndarray, optional):
                Posterior probability. If not specified, ``posterior`` is computed by current
                parameters.

        Returns:
            numpy.ndarray of the separated signal in frequency-domain.
            The shape is (n_sources, n_bins, n_frames).
        """
        X = input

        if posterior is None:
            alpha = self.mixing
            Z = self.unit_input
            B = self.covariance

            Z = Z.transpose(1, 2, 0)
            B_inverse = np.linalg.inv(B)
            ZBZ = quadratic(Z, B_inverse[:, :, np.newaxis])
            ZBZ = np.real(ZBZ)
            ZBZ = np.maximum(ZBZ, 0)
            ZBZ = self.flooring_fn(ZBZ)

            log_alpha = np.log(alpha)
            _, logdet = np.linalg.slogdet(B)
            log_prob = log_alpha - logdet
            log_gamma = log_prob[:, :, np.newaxis] - self.n_channels * np.log(ZBZ)

            gamma = softmax(log_gamma, axis=0)
        else:
            gamma = posterior

        return gamma * X[self.reference_id]

    def update_once(
        self, flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self"
    ) -> None:
        r"""Perform E and M step once.

        In ``update_posterior``, posterior probabilities are updated, which corresponds to E step.
        In ``update_parameters``, parameters of cACGMM are updated, which corresponds to M step.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        """
        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        self.update_posterior(flooring_fn=flooring_fn)
        self.update_parameters(flooring_fn=flooring_fn)

        if self.normalization:
            self.normalize_covariance()

    def update_posterior(
        self, flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self"
    ) -> None:
        r"""Update posteriors.

        This method corresponds to E step in EM algorithm for cACGMM.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        """
        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        alpha = self.mixing
        Z = self.unit_input
        B = self.covariance

        Z = Z.transpose(1, 2, 0)
        B_inverse = np.linalg.inv(B)
        ZBZ = quadratic(Z, B_inverse[:, :, np.newaxis])
        ZBZ = np.real(ZBZ)
        ZBZ = np.maximum(ZBZ, 0)
        ZBZ = flooring_fn(ZBZ)

        log_prob = np.log(alpha) - self.compute_logdet(B)
        log_gamma = log_prob[:, :, np.newaxis] - self.n_channels * np.log(ZBZ)

        gamma = softmax(log_gamma, axis=0)

        self.posterior = gamma

    def update_parameters(
        self, flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self"
    ) -> None:
        r"""Update parameters of mixture of complex angular central Gaussian distributions.

        This method corresponds to M step in EM algorithm for cACGMM.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.
        """
        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        Z = self.unit_input
        B = self.covariance
        gamma = self.posterior

        Z = Z.transpose(1, 2, 0)
        B_inverse = np.linalg.inv(B)
        ZBZ = quadratic(Z, B_inverse[:, :, np.newaxis])
        ZBZ = np.real(ZBZ)
        ZBZ = np.maximum(ZBZ, 0)
        ZBZ = flooring_fn(ZBZ)
        ZZ = Z[:, :, :, np.newaxis] * Z[:, :, np.newaxis, :].conj()

        alpha = np.mean(gamma, axis=-1)

        GZBZ = gamma / ZBZ
        num = np.sum(GZBZ[:, :, :, np.newaxis, np.newaxis] * ZZ, axis=2)
        denom = np.sum(gamma, axis=2)
        B = self.n_channels * (num / denom[:, :, np.newaxis, np.newaxis])
        B = to_psd(B, flooring_fn=flooring_fn)

        self.mixing = alpha
        self.covariance = B

    def compute_loss(self) -> float:
        r"""Compute loss of cACGMM :math:`\mathcal{L}`.

        :math:`\mathcal{L}` is defined as follows:

        .. math::
            \mathcal{L}
            = -\frac{1}{J}\sum_{i,j}\log\left(
            \sum_{n}\frac{\alpha_{in}}{\det\boldsymbol{B}_{in}}
            \frac{1}{(\boldsymbol{z}_{ij}^{\mathsf{H}}\boldsymbol{B}_{in}^{-1}\boldsymbol{z}_{ij})^{M}}
            \right).
        """
        alpha = self.mixing
        Z = self.unit_input
        B = self.covariance

        Z = Z.transpose(1, 2, 0)
        B_inverse = np.linalg.inv(B)
        ZBZ = quadratic(Z, B_inverse[:, :, np.newaxis])
        ZBZ = np.real(ZBZ)
        ZBZ = np.maximum(ZBZ, 0)
        ZBZ = self.flooring_fn(ZBZ)

        log_prob = np.log(alpha) - self.compute_logdet(B)
        log_gamma = log_prob[:, :, np.newaxis] - self.n_channels * np.log(ZBZ)

        loss = -logsumexp(log_gamma, axis=0)
        loss = np.mean(loss, axis=-1)
        loss = loss.sum(axis=0)
        loss = loss.item()

        return loss



import warnings
from typing import Callable, List, Optional, Union

import numpy as np

from ..linalg import prox
from .proxbss import ProxBSSBase

EPS = 1e-10


class ADMMBSSBase(ProxBSSBase):
    """Base class of blind source separation via alternative direction method of multiplier.

    Args:
        penalty_fn (callable):
            Penalty function that determines source model.
        prox_penalty (callable):
            Proximal operator of penalty function.
            Default: ``None``.
        callbacks (callable or list[callable], optional):
            Callback functions. Each function is called before separation and at each iteration.
            Default: ``None``.
        scale_restoration (bool or str):
            Technique to restore scale ambiguity.
            If ``scale_restoration=True``, the projection back technique is applied to
            estimated spectrograms. You can also specify ``projection_back`` explicitly.
            Default: ``True``.
        record_loss (bool):
            Record the loss at each iteration of the update algorithm if ``record_loss=True``.
            Default: ``True``.
        reference_id (int):
            Reference channel for projection back.
            Default: ``0``.

    """

    def __repr__(self) -> str:
        s = "ADMMBSS("
        s += "n_penalties={n_penalties}".format(n_penalties=self.n_penalties)
        s += ", scale_restoration={scale_restoration}"
        s += ", record_loss={record_loss}"

        if self.scale_restoration:
            s += ", reference_id={reference_id}"

        s += ")"

        return s.format(**self.__dict__)


class ADMMBSS(ADMMBSSBase):
    """Base class of blind source separation via alternative direction method of multiplier.

    Args:
        rho (float):
            Penalty parameter. Default: ``1``.
        alpha (float):
            Relaxation parameter (deprecated). Set ``relaxation`` instead.
        relaxation (float):
            Relaxation parameter. Default: ``1``.
        penalty_fn (callable):
            Penalty function that determines source model.
        prox_penalty (callable):
            Proximal operator of penalty function.
            Default: ``None``.
        callbacks (callable or list[callable], optional):
            Callback functions. Each function is called before separation and at each iteration.
            Default: ``None``.
        scale_restoration (bool or str):
            Technique to restore scale ambiguity.
            If ``scale_restoration=True``, the projection back technique is applied to
            estimated spectrograms. You can also specify ``projection_back`` explicitly.
            Default: ``True``.
        record_loss (bool):
            Record the loss at each iteration of the update algorithm if ``record_loss=True``.
            Default: ``True``.
        reference_id (int):
            Reference channel for projection back.
            Default: ``0``.

    """

    def __init__(
        self,
        rho: float = 1,
        alpha: float = None,
        relaxation: float = 1,
        penalty_fn: Callable[[np.ndarray, np.ndarray], float] = None,
        prox_penalty: Callable[[np.ndarray, float], np.ndarray] = None,
        callbacks: Optional[
            Union[Callable[["ADMMBSS"], None], List[Callable[["ADMMBSS"], None]]]
        ] = None,
        scale_restoration: bool = True,
        record_loss: bool = True,
        reference_id: int = 0,
    ) -> None:
        super().__init__(
            penalty_fn=penalty_fn,
            prox_penalty=prox_penalty,
            callbacks=callbacks,
            scale_restoration=scale_restoration,
            record_loss=record_loss,
            reference_id=reference_id,
        )

        self.rho = rho

        if alpha is None:
            self.relaxation = relaxation
        else:
            assert relaxation == 1, "You cannot specify relaxation and alpha simultaneously."

            warnings.warn("alpha is deprecated. Set relaxation instead.", DeprecationWarning)

            self.relaxation = alpha

    def __call__(self, input, n_iter=100, initial_call: bool = True, **kwargs) -> np.ndarray:
        r"""Separate a frequency-domain multichannel signal.

        Args:
            input (numpy.ndarray):
                Mixture signal in frequency-domain.
                The shape is (n_channels, n_bins, n_frames).
            n_iter (int):
                Number of iterations of demixing filter updates.
                Default: ``100``.
            initial_call (bool):
                If ``True``, perform callbacks (and computation of loss if necessary)
                before iterations.

        Returns:
            numpy.ndarray of the separated signal in frequency-domain.
            The shape is (n_channels, n_bins, n_frames).
        """
        self.input = input.copy()

        self._reset(**kwargs)

        # Call __call__ of ADMMBSSBase's parent, i.e. __call__ of IterativeMethodBase
        super(ADMMBSSBase, self).__call__(n_iter=n_iter, initial_call=initial_call)

        if self.scale_restoration:
            self.restore_scale()

        self.output = self.separate(self.input, demix_filter=self.demix_filter)

        return self.output

    def __repr__(self) -> str:
        s = "ADMMBSS("
        s += "rho={rho}"
        s += ", relaxation={relaxation}"
        s += ", n_penalties={n_penalties}".format(n_penalties=self.n_penalties)
        s += ", scale_restoration={scale_restoration}"
        s += ", record_loss={record_loss}"

        if self.scale_restoration:
            s += ", reference_id={reference_id}"

        s += ")"

        return s.format(**self.__dict__)

    def _reset(self, **kwargs) -> None:
        r"""Reset attributes by given keyword arguments.

        Args:
            kwargs:
                Keyword arguments to set as attributes of ADMMBSS.
        """
        super()._reset(**kwargs)

        n_penalties = self.n_penalties
        n_sources, n_channels = self.n_sources, self.n_channels
        n_bins, n_frames = self.n_bins, self.n_frames

        if not hasattr(self, "aux1"):
            aux1 = np.zeros((n_bins, n_sources, n_channels), dtype=np.complex128)
        else:
            if self.aux1 is None:
                aux1 = None
            else:
                # To avoid overwriting ``aux1`` given by keyword arguments.
                aux1 = self.aux1.copy()

        if not hasattr(self, "aux2"):
            aux2 = np.zeros((n_penalties, n_sources, n_bins, n_frames), dtype=np.complex128)
        else:
            if self.aux2 is None:
                aux2 = None
            else:
                # To avoid overwriting ``aux2`` given by keyword arguments.
                aux2 = self.aux2.copy()

        if not hasattr(self, "dual1"):
            dual1 = np.zeros((n_bins, n_sources, n_channels), dtype=np.complex128)
        else:
            if self.dual1 is None:
                dual1 = None
            else:
                # To avoid overwriting ``dual1`` given by keyword arguments.
                dual1 = self.dual1.copy()

        if not hasattr(self, "dual2"):
            dual2 = np.zeros((n_penalties, n_sources, n_bins, n_frames), dtype=np.complex128)
        else:
            if self.dual2 is None:
                dual2 = None
            else:
                # To avoid overwriting ``dual2`` given by keyword arguments.
                dual2 = self.dual2.copy()

        self.aux1 = aux1
        self.aux2 = aux2
        self.dual1 = dual1
        self.dual2 = dual2

    def update_once(self) -> None:
        r"""Update demixing filters, auxiliary parameters, and dual parameters once."""
        n_penalties = self.n_penalties
        n_channels = self.n_channels
        rho, alpha = self.rho, self.relaxation

        V, V_tilde = self.aux1, self.aux2
        Y, Y_tilde = self.dual1, self.dual2
        X, W = self.input, self.demix_filter

        XX = X.transpose(1, 0, 2).conj() @ X.transpose(1, 2, 0)
        E = np.eye(n_channels)
        VY = V - Y
        VY_tilde = np.sum(V_tilde - Y_tilde, axis=0)
        XVY_tilde = X.transpose(1, 0, 2).conj() @ VY_tilde.transpose(1, 2, 0)

        W = np.linalg.solve(n_penalties * XX + E, VY + XVY_tilde.transpose(0, 2, 1))
        XW = self.separate(X, demix_filter=W)

        U = alpha * W + (1 - alpha) * V
        U_tilde = alpha * XW + (1 - alpha) * V_tilde

        V = prox.neg_logdet(U + Y, step_size=1 / rho)

        V_tilde = []

        for U_tilde_q, Y_tilde_q, prox_penalty in zip(U_tilde, Y_tilde, self.prox_penalty):
            V_tilde_q = prox_penalty(U_tilde_q + Y_tilde_q, step_size=1 / rho)
            V_tilde.append(V_tilde_q)

        V_tilde = np.stack(V_tilde, axis=0)

        Y = Y + U - V
        Y_tilde = Y_tilde + U_tilde - V_tilde

        self.aux1, self.aux2 = V, V_tilde
        self.dual1, self.dual2 = Y, Y_tilde
        self.demix_filter = W
